#!/bin/bash

# Copyright (c) 2019-2022, NVIDIA CORPORATION.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# This script generates OTA package, which used to ota from
# R28.2/R28.3/R28.4/R32.1/R32.2/R32.3/R32.4/R32.5/R32.6/R32.7 to R32-ToT.
set -e

LINUX_BASE_DIR="$(pwd)"
OTA_DIR="${LINUX_BASE_DIR}/tools/ota_tools/version_upgrade"
BOOTLOADER_DIR="${LINUX_BASE_DIR}/bootloader"
OTA_BASE_DIR_TMP=""
SYSTEM_IMAGE_RAW_FILE="${BOOTLOADER_DIR}/system.img.raw"
SYSTEM_IMAGE_TMP=""
OTA_PACKAGE_FILE="ota_package.tar"
OTA_PACKAGE_SHA1_FILE="ota_package.tar.sha1sum"
OTA_PAYLOAD_PACKAGE_ZIP_FILE="ota_payload_package.tar.gz"
R32_MB1_FILE="mb1_t194_prod_r32.bin"
DUMMY_SMD_FILE="dummy.bin"
DUMMY_SMD_FILE_SIZE=191

# Do not use underscore "_" in these dir name
R35_TOT_IMAGES_DIR="images-R35-ToT"
R32x_R35i_IMAGES_DIR="images-R32x-R35i"
R35A_R35i_IMAGES_DIR="images-R35A-R35i"
R32i_IMAGES_DIR="images-R32i"

COMMON_TAR_OPTIONS=( --warning=none --checkpoint=10000 --one-file-system . )
PKC_KEY_FILE=""
SBK_KEY_FILE=""
BASE_VERSION=""
TARGET_BOARD=""
# Skip re-generating recovery image if it exists
SKIP_REC_IMG=0
BOARD_SPECS_ARRAY=""
CHIPID=""
# Images on following partitions are board specific.
# The image for recovery partition is a special case. It is not
# board specific, but the recovery image is dynamically generated
# and its content is changed each time the command for generating
# images is executed. So need to put it here to make sure that
# the recovery image can match the corresponding "flash.idx".
BOARD_DEP_PARTS=( "BCT" "MB1_BCT" "bootloader-dtb" "bpmp-fw-dtb" "kernel-dtb" "recovery-dtb" "VER" "esp" )

# kernel/kernerl-dtb/recovery/recovery-dtb images are not signed
# but images for kernel-dtb/recovery-dtb are board-specific, they
# will be copied into board-specific directory.
UNSIGNED_PARTS=( "kernel" "recovery" )

# Support generating OTA payload package for the case
# with/without partition layout change
LAYOUT_CHANGE=0
# Utilities needed by doing BUP update.
BUP_UPDATE_UTILITIES=( "nv_update_engine" "nv_bootloader_payload_updater" "nvbootctrl" )
# Utilities needed for partition update.
PARTITION_UPDATE_UTILITIES=( "nvbootctrl" )
# Support for specifying the rootfs update script by user
NV_ROOTFS_UPDATER="nv_ota_rootfs_updater.sh"
# Customer configuration file for image-based OTA
OTA_CUSTOMER_CONF="${OTA_DIR}/nv_ota_customer.conf"

function usage()
{
	echo -ne "Usage: sudo $0 [options] <target board> <bsp version>\n"
	echo -ne "\tWhere,\n"
	echo -ne "\t\t<target board>: target board. Supported boards: jetson-agx-xavier-devkit, jetson-xavier-nx-devkit-emmc.\n"
	echo -ne "\t\t<bsp version>: the version of the base BSP. Supported versions: R32-7.\n"
	echo -ne "\toptions:\n"
	echo -ne "\t\t-u <PKC key file>: PKC key used for odm fused board\n"
	echo -ne "\t\t-v <SBK key file>: Secure Boot Key (SBK) key used for ODM fused board\n"
	echo -ne "\t\t-s\tSkip generating system image.\n"
	echo -ne "\t\t-b\tUpdate bootloader only. Only valid for update without layout change.\n"
	echo -ne "\t\t-r\tUpdate rootfs only. Only valid for update without layout change.\n"
	echo -ne "\t\t-o\tSpecify the script to update rootfs partition.\n"
	echo -ne "\t\t-f\tSpecify the rootfs image to be written to rootfs partition.\n"
	echo -ne "Example:\n"
	echo -ne "\tsudo $0 -u PKCkeyfile -v SBKkeyfile jetson-xavier-nx-devkit-emmc R32-7\n"
	exit 1
}

function check_intermediate_partitions()
{
	local a_idx_file="${1}"
	local b_idx_file="${2}"

	mapfile -t a_msi_partitions < <(awk '/MSI_EMMC_START/, EOF {print $2 $3}' "${a_idx_file}")
	mapfile -t b_msi_partitions < <(awk '/MSI_EMMC_START/, EOF {print $2 $3}' "${b_idx_file}")
	if [ "${#a_msi_partitions[@]}" != "${#b_msi_partitions[@]}" ]; then
		echo "Error: the MSI partitions layout in ${a_idx_file} and ${b_idx_file} do not match"
		exit 1
	fi

	# Compared each partition of MSI in the 2 intermediate layout file and
	# make sure the offset of each of them is not changed. If the offset changes,
	# some partition might be missed and OTA process might be failed.
	mapfile -t diff < <(paste -d, <(printf '%s\n' "${a_msi_partitions[@]}") <(printf '%s\n' "${b_msi_partitions[@]}") | awk -F, '$2 != $5')
	if [ "${#diff[@]}" -gt 0 ]; then
		for item in "${diff[@]}"
		do
			IFS=, read -r a_part a_part_offset _ b_part b_part_offset <<< "${item}"
			echo "Error: ${a_part##*:}:${a_part_offset} does not match with ${b_part##*:}:${b_part_offset}"
		done
		exit 1
	fi
}

function check_bootchain_a_partitions()
{
	local current_idx_file="${1}"
	local next_idx_file="${2}"

	# Find all bootchain A partition names in the boot device
	mapfile -t b_boot_part_names < \
		<(awk '$2 ~ NR==1,/secondary_gpt/ {if($2 ~ /.*_b,$/ && $2 !~ /VER/) print $2}' "${current_idx_file}" | sed -r 's/(.*:)(.*)(,$)/\2/')
	local a_boot_part_names=("${b_boot_part_names[@]/_b/}")

	# Pick up bootchain A partitions from the current layout
	mapfile -t current_parts < <(awk '{print $2 $3}' "${current_idx_file}")
	local a_boot_parts=()
	for part in "${current_parts[@]}"
	do
		IFS=, read -r name _ <<< "${part}"
		if [[ " ${a_boot_part_names[*]} " == *" ${name##*:} "* ]]; then
			a_boot_parts+=("${part}")
		fi
	done

	# Compare the offset of each bootchain A partition with that in the next
	# intermediate layout.
	for part in "${a_boot_parts[@]}"
	do
		IFS=, read -r name offset <<< "${part}"
		_offset=$(awk -v pattern="^${name},$" '$2 ~ pattern {sub(",", "", $3); print $3}' "${next_idx_file}")
		if [ "${offset}" != "${_offset}" ]; then
			echo "Error: ${name##*:}: ${offset} does not match with ${_offset}"
			exit 1
		fi
	done
}

function init()
{
	OTA_BASE_DIR_TMP="${LINUX_BASE_DIR}/R32x_R35-ToT_OTA"
	SYSTEM_IMAGE_TMP="${OTA_BASE_DIR_TMP}/sysimg_tmp"
	if [ -e "${OTA_BASE_DIR_TMP}" ]; then
		rm "${OTA_BASE_DIR_TMP}" -rf
	fi
	mkdir "${OTA_BASE_DIR_TMP}"
}

function construct_board_spec_name()
{
	# Construct the board spec name based on the items:
	# boardid, fab, boardsku and boardrev
	# Usage:
	#	construct_board_spec_name {return_value}
	local return_value="$1"
	local name=
	local l_boardid=
	local l_fab=
	local l_boardsku=
	local l_boardrev=

	l_boardid="${boardid}"
	l_fab="${fab}"
	l_boardsku="${boardsku}"
	l_boardrev="${boardrev}"

	# The boardid and fab are compulsory, so return error
	# if either of them is empty.
	if [ "${l_boardid}" == "" ] || [ "${l_fab}" == "" ]; then
		echo "Both BOARDID and FAB must be set"
		exit 1
	fi

	# Board spec name is composed of boardid, fab, boardsku and
	# boardrev, like this:
	# board_spec_name=${l_boardid}-${l_fab}-${l_boardsku}-${l_boardrev}
	name="${l_boardid}-${l_fab}-${l_boardsku}-${l_boardrev}"
	eval "${return_value}=${name}"
}

function copy_board_dep_files()
{
	# Copy the files for some compatible board spec into the
	# sub-directory named as the compatible board spec name under
	# the "images-XXX-XXX"
	# Usage:
	#	 copy_board_dep_files {src_dir} {dest_dir}
	local src_dir="$1"
	local dest_dir="$2"
	local idx_file="${src_dir}/flash.idx"
	local partition=
	local file_name=

	# Copy the images dependent on board SPEC
	for partition in "${BOARD_DEP_PARTS[@]}"; do
		file_name="$(grep ":${partition}," "${idx_file}" | cut -d, -f 5 | sed 's/^ //' -)"
		if [ "${file_name}" == "" ]; then
			# For the intermedate layout file, the file for the "${partition}"
			# might be empty, then try seach the "${partition}_b"
			file_name="$(grep ":${partition}_b," "${idx_file}" | cut -d, -f 5 | sed 's/^ //' -)"
			if [ "${file_name}" == "" ]; then
				echo "Warning: the file written to partition ${partition} is not found"
				continue
			fi
		fi

		local file_list=("${file_name}")
		if [ "${partition}" == "BCT" ]; then
			file_list+=("${file_name/br_bct/br_bct_b}")
		fi

		for file_name in "${file_list[@]}"
		do
			if [ -f "${src_dir}/${file_name}" ]; then
				cp "${src_dir}/${file_name}" "${dest_dir}/"
			else
				# For some files, such as file for VER/VER_b partition, it is
				# not signed, so it should be under the parent directory of "src_dir"
				if [ -f "${src_dir}/../${file_name}" ]; then
					cp "${src_dir}/../${file_name}" "${dest_dir}/"
				else
					echo "Error: the file ${src_dir}/../${file_name} is not found"
					return 1
				fi
			fi
		done
	done

	# copy images that are dynamically generated each time the command
	# for generating images is executed into the sub-directories.
	# The content of these images are different and their sha1sum are
	# different in the "flash.idx".
	# These images are for partitions: secondary_gpt, primary_gpt
	cp "${src_dir}"/gpt*.bin "${dest_dir}"/

	# Copy the flash.idx file
	cp "${idx_file}" "${dest_dir}/"

	return 0
}

function copy_signed_files()
{
	# Copy the files from the directory storing all signed images
	# into the "images-XXX-XXX" directory and its sub-directories
	# which are named as compatible spec name
	# Usage:
	#	 copy_signed_files {dest_dir}
	local dest_dir="$1"
	local src_dir=

	# If SBK is non-zero, the encryted/signed files are under "encrypted_signed",
	# else, the signed files are under "signed"
	if [ -s "${SBK_KEY_FILE}" ]; then
		echo "encrypted_signed_dir=${dest_dir}"
		if [ "${CHIPID}" == "0x19" ]; then
			src_dir="${BOOTLOADER_DIR}/encrypted_signed_t19x"
		else
			src_dir="${BOOTLOADER_DIR}/encrypted_signed"
		fi
	else
		echo "signed_dir=${dest_dir}"
		src_dir="${BOOTLOADER_DIR}/signed"
	fi

	# copy the files dpendent on the board specification into the
	# directory whose name is corresponding compatible spec name.
	if ! copy_board_dep_files "${src_dir}" "${dest_dir}"; then
		echo "Error: failed copy boardspec dependent files from ${src_dir} to ${dest_dir}"
		return 1
	fi

	# copy all the files into the "images-XXX-XXX" directory
	dest_dir="${dest_dir%/*}"
	cp -r "${src_dir}"/* "${dest_dir}"/

	return 0
}

function copy_unsigned_files()
{
	local dest_dir="$1"
	local src_dir="${BOOTLOADER_DIR}"
	local idx_file="${src_dir}/signed/flash.idx"

	for partition in "${UNSIGNED_PARTS[@]}"
	do
		file_name="$(grep ":${partition}," "${idx_file}" | cut -d, -f 5 | sed 's/^ //' -)"
		if [ "${file_name}" == "" ]; then
			continue
		fi
		if [ ! -f "${src_dir}/${file_name}" ]; then
			echo "Error: ${src_dir}/${file_name} is not found"
			return 1
		fi
		cp "${src_dir}/${file_name}" "${dest_dir}"
	done
	return 0
}

function generate_binaries()
{
	local ret_msg=""
	local src_dir=""
	local dest_dir=""
	local spec=
	local board_specs=
	local fuselevel="fuselevel_production"
	local board_spec_name=
	local system_img_built=0

	if [ -z "${BASE_BSP}" ]; then
		echo "Error: \"BASE_BSP\" is not set"
		exit 1
	fi
	echo "Copying mb1 image from ${BASE_BSP} for re-signing"
	cp -f "${BASE_BSP}/bootloader/${R32_MB1_FILE/_r32/}" "${BOOTLOADER_DIR}/${R32_MB1_FILE}"

	# Generate dummy SMD image for SMD partitions in the intermediate layouts
	dd if=/dev/zero of="${BOOTLOADER_DIR}/${DUMMY_SMD_FILE}" bs=1 count="${DUMMY_SMD_FILE_SIZE}"

	# Remove existing "${SYSTEM_IMAGE_RAW_FILE}" if "{skip_system_img}" is 0
	if  [ -f "${SYSTEM_IMAGE_RAW_FILE}" ] && [ "${skip_system_img}" == 0 ] ; then
		rm -f "${SYSTEM_IMAGE_RAW_FILE}"
	fi

	board_specs="${!BOARD_SPECS_ARRAY}[@]"
	for spec in "${!board_specs}"; do
		eval "${spec}"

		# Report error if "signed_img_dir" is not set in "ota_board_specs.conf"
		if [ "${signed_img_dir}" == "" ]; then
			echo "ERROR: the \"signed_img_dir\" is not set"
			exit 1
		fi

		board_arg="BOARDID=${boardid} FAB=${fab} BOARDSKU=${boardsku} "
		board_arg+="BOARDREV=${boardrev} FUSELEVEL=${fuselevel}"

		cmd_arg="--no-flash --sign "
		if [ "${PKC_KEY_FILE}" != "" ] && [ -f "${PKC_KEY_FILE}" ]; then
			cmd_arg+="-u \"${PKC_KEY_FILE}\" "
		fi
		if [ "${SBK_KEY_FILE}" != "" ] && [ -f "${SBK_KEY_FILE}" ]; then
			cmd_arg+="-v \"${SBK_KEY_FILE}\" "
		fi
		# Do not generate system.img.raw if
		# 1) Updating bootloader only.
		# 2) Skip generating system image
		# 3) User specified rootfs image is set
		if [ "${update_bl_only}" == 1 ] || \
			[ "${skip_system_img}" == 1 ] || \
			[ -n "${rootfs_image}" ] || \
			[ "${system_img_built}" == 1 ]; then
			sysimg="--no-systemimg"
		else
			sysimg=""
			system_img_built=1
		fi
		cmd_arg+="${sysimg} "
		cmd_arg+="${board} ${rootdev}"

		cmd="SKIP_REC_IMG=${SKIP_REC_IMG} ${board_arg} ${LINUX_BASE_DIR}/flash.sh ${cmd_arg}"

		# construct the name of the sub-directory by parsing the board spec
		if ! construct_board_spec_name "board_spec_name"; then
			echo "Failed to construct board spec name"
			exit 1
		fi

		# generate R35 binaries
		echo -e "${cmd}\r\n"
		if eval "${cmd}"; then
			# Create destination directory
			dest_dir="${OTA_BASE_DIR_TMP}/${signed_img_dir}/${board_spec_name}"
			mkdir -p "${dest_dir}"

			# Copy the signed files into destination directory
			if ! copy_signed_files "${dest_dir}"; then
				echo "Error: failed to copy signed files in the ${dest_dir}"
				exit 1
			fi
			# Copy the unsigned files into "images-XXX-XXX" directory
			if ! copy_unsigned_files "${dest_dir%/*}"; then
				echo "Error: failed to copy unsigned files in the ${dest_dir}"
				exit 1
			fi
			ret_msg+="\r\nSUCCESS: generate binaries for OTA \"${spec}\""
		else
			ret_msg+="\r\nFAILURE: no binaires for OTA \"${spec}\""
			exit 1
		fi

		# Generate recovery image only one time as its content never changes
		# during generating images for the same board. No need to re-generate
		# it every time calling the "flash.sh" to generate images.
		SKIP_REC_IMG=1

		# Copy the "/etc/nv_boot_control.conf" from rootfs to ${OTA_BASE_DIR_TMP}/.
		# When generating images for intermediate layout, the "/etc/nv_boot_control.conf"
		# in the rootfs is to be updated with the corresponding board name, such as
		# "jetson-xavier-R32A-R32i-emmc", but this board name is not expected.
		# The expected name should equals to the ${TARGET_BOARD}, so need to copy
		# the "/etc/nv_boot_control.conf" from rootfs after successfully generating
		# images with the "${board}" equals "${TARGET_BOARD}"
		if [ "${board}" == "${TARGET_BOARD}" ]; then
			local rootfs_dir="${LINUX_BASE_DIR}/rootfs"
			local nv_boot_control_conf="${rootfs_dir}/etc/nv_boot_control.conf"
			local ota_nv_boot_control_conf="${OTA_DIR}/ota_nv_boot_control.conf"
			if [ ! -f "${nv_boot_control_conf}" ]; then
				echo "The file ${nv_boot_control_conf} is not found"
				exit 1
			fi
			cp "${nv_boot_control_conf}" "${ota_nv_boot_control_conf}"
		fi

		# For no layout change case, only recovery image is needed, so the loop can exit
		# once the flash.sh script is executed one time. Then the generated images are stored
		# in the "images-R32-ToT" directory.
		if [ "${LAYOUT_CHANGE}" == 0 ]; then
			echo -e "${ret_msg}"
			return 0
		fi
	done

	# Check the 2 intermedate layout files to make sure that the offset of
	# the partitions are matched.
	read -r first_msi_interm_dir second_msi_interm_dir <<< "$(get_msi_interm_dirs "${OTA_BASE_DIR_TMP}")"
	local first_msi_interm_idx_file="${first_msi_interm_dir}/${board_spec_name}/flash.idx"
	local second_msi_interm_idx_file="${second_msi_interm_dir}/${board_spec_name}/flash.idx"
	check_intermediate_partitions "${first_msi_interm_idx_file}" "${second_msi_interm_idx_file}"

	# Special handling for platform like Xavier NX which uses an extra intermediate
	# layout R32i as the first phrase. Need to make sure bootchain a partitions are
	# matched with that in the next intermediate layout.
	local first_interm_dir=
	first_interm_dir=$(get_first_interm_dir "${OTA_BASE_DIR_TMP}")
	local first_interm_idx_file="${first_interm_dir}/${board_spec_name}/flash.idx"
	if [ "${first_interm_idx_file}" != "${first_msi_interm_idx_file}" ]; then
		check_bootchain_a_partitions "${first_interm_idx_file}" "${first_msi_interm_idx_file}"
	fi

	rm -f "${BOOTLOADER_DIR}/${R32_MB1_FILE}"
	rm -f "${BOOTLOADER_DIR}/${DUMMY_SMD_FILE}"
	echo -e "${ret_msg}"
}

function tar_systemimg()
{
	local app_system_file="${OTA_BASE_DIR_TMP}/system.img"
	local app_system_sha1_file="${OTA_BASE_DIR_TMP}/system.img.sha1sum"
	local sha1_chksum=

	# If user specifies his own rootfs image and corresponding rootfs updater,
	# copy the specified rootfs image into "${OTA_BAS_DIR_TMP}" directly
	# If user specifies his own rootfs image, but not specify corresponding
	# rootfs updater, replace "system.img.raw" with user's rootfs image.
	if [ -n "${rootfs_image}" ]; then
		if [ "${rootfs_updater}" != "" ]; then
			cp -f "${rootfs_image}" "${app_system_file}"
			sha1_chksum="$(sha1sum "${app_system_file}" | cut -d\  -f 1)"
			echo -n "${sha1_chksum}" > "${app_system_sha1_file}"
			return 0
		else
			cp -f "${rootfs_image}" "${SYSTEM_IMAGE_RAW_FILE}"
		fi
	fi
	mkdir "${SYSTEM_IMAGE_TMP}"

	# Mount system.img.raw"
	if ! mount "${SYSTEM_IMAGE_RAW_FILE}" "${SYSTEM_IMAGE_TMP}"; then
		echo "Failed to mount ${SYSTEM_IMAGE_RAW_FILE}"
		rm "${SYSTEM_IMAGE_TMP}" -rf
		exit 1
	fi

	pushd "${SYSTEM_IMAGE_TMP}" > /dev/null 2>&1
	set +e

	# Making APP image
	if ! tar cpf - "${COMMON_TAR_OPTIONS[@]}" | gzip -1 > "${app_system_file}"; then
		echo "Failed to make APP image"
		popd > /dev/null 2>&1
		rm "${SYSTEM_IMAGE_TMP}" -rf
		exit 1
	fi

	# Generate sha1 checksum for the app image
	sha1_chksum="$(sha1sum "${app_system_file}" | cut -d\  -f 1)"
	echo -n "${sha1_chksum}" > "${app_system_sha1_file}"

	set -e
	popd > /dev/null 2>&1
	umount "${SYSTEM_IMAGE_TMP}"
	rm "${SYSTEM_IMAGE_TMP}" -rf

	echo -e "\r\nSUCCESS: generate APP image"
}

function copy_ota_tasks()
{
	# Load variable OTA_TASKS and OTA_TASK_NEEDED_FILES defined in nv_ota_customer.conf
	local _declare=
	_declare=$(source "${OTA_CUSTOMER_CONF}" && declare -p OTA_TASKS OTA_TASK_NEEDED_FILES)
	eval "${_declare}"

	if [ -z "${OTA_TASKS}" ]; then
		echo "Error: can not find OTA_TASKS in ${OTA_CUSTOMER_CONF}"
		exit 1
	fi

	pushd "${OTA_DIR}"
	cp -v "${OTA_TASKS[@]}" "${OTA_TASK_NEEDED_FILES[@]}" "${OTA_BASE_DIR_TMP}"/
	popd
}

function copy_common_images()
{
	if [[ ! "${BASE_VERSION}" =~ "R32-7" ]]; then
		local version="${1}"
		local base_recovery_img=
		local base_recovery_img_sha1=
		local base_recovery_dtb=
		local base_recovery_dtb_sha1=
		base_recovery_img="recovery.img.${version}"
		base_recovery_img_sha1="recovery.img.${version}.sha1sum"
		base_recovery_dtb="recovery.dtb.${version}"
		base_recovery_dtb_sha1="recovery.dtb.${version}.sha1sum"

		# Copy recovery images
		cp "${BOOTLOADER_DIR}/${base_recovery_img}" "${OTA_BASE_DIR_TMP}"
		cp "${BOOTLOADER_DIR}/${base_recovery_img_sha1}" "${OTA_BASE_DIR_TMP}"
		cp "${BOOTLOADER_DIR}/${base_recovery_dtb}" "${OTA_BASE_DIR_TMP}"
		cp "${BOOTLOADER_DIR}/${base_recovery_dtb_sha1}" "${OTA_BASE_DIR_TMP}"
	fi

	# Generate base version file and board name file
	echo -n "${TARGET_BOARD}" >"${OTA_BASE_DIR_TMP}/board_name"
	echo -n "${BASE_VERSION}" >"${OTA_BASE_DIR_TMP}/base_version"

	# Pack "nv_ota_preserve_data.sh" into the OTA payload package.
	local nv_ota_preserve_data="nv_ota_preserve_data.sh"
	cp "${OTA_DIR}"/"${nv_ota_preserve_data}" "${OTA_BASE_DIR_TMP}"/

	# Copy NVIDIA provided rootfs updater to ${OTA_BASE_DIR_TMP}
	local nv_rootfs_updater="${OTA_DIR}/${NV_ROOTFS_UPDATER}"
	echo "Copy NVIDIA provided rootfs updater to ${OTA_BASE_DIR_TMP}/${NV_ROOTFS_UPDATER}"
	cp "${nv_rootfs_updater}" "${OTA_BASE_DIR_TMP}"/"${NV_ROOTFS_UPDATER}"

	# Copy customer provided rootfs updater
	if [ "${rootfs_updater}" != "" ]; then
		echo "Copy customer provided rootfs updater to ${OTA_BASE_DIR_TMP}/${rootfs_updater_name}"
		cp "${rootfs_updater}" "${OTA_BASE_DIR_TMP}"/"${rootfs_updater_name}"
	fi

	# Pack "nv_ota_run_tasks.sh" into the OTA payload package.
	local nv_ota_run_tasks="nv_ota_run_tasks.sh"
	cp "${OTA_DIR}"/"${nv_ota_run_tasks}" "${OTA_BASE_DIR_TMP}"/

	# Copy nv_ota_customer.conf to ${OTA_BASE_DIR_TMP}
	cp "${OTA_CUSTOMER_CONF}" "${OTA_BASE_DIR_TMP}"/

	# Copy kernel_bootctrl binaries
	binary_list=(
		"${OTA_DIR}/kernel_bootctrl.bin.normal"
		"${OTA_DIR}/kernel_bootctrl.bin.update"
	)
	for bin in "${binary_list[@]}"; do
		cp "${bin}" "${OTA_BASE_DIR_TMP}"
	done

	return 0
}

function copy_nv_bootloader_config_files()
{
	# Copy the "/opt/nvidia/l4t-bootloader-config/nv-l4t-bootloader-config.sh"
	# from rootfs to ${OTA_BASE_DIR_TMP}/.
	# For supporting multiple board spec, the "nv-l4t-bootloader-config.sh"
	# and the "${OTA_BASE_DIR_TMP}/ota_nv_boot_control.conf" are used to
	# update the "/etc/nv_boot_contro.conf".
	local rootfs_dir="${LINUX_BASE_DIR}/rootfs"
	local update_boot_control_script=
	update_boot_control_script="${rootfs_dir}/opt/nvidia/l4t-bootloader-config/nv-l4t-bootloader-config.sh"
	if [ ! -f "${update_boot_control_script}" ]; then
		echo "The file ${update_boot_control_script} is not found"
		exit 1
	fi
	cp "${update_boot_control_script}" "${OTA_BASE_DIR_TMP}"/
	cp "${OTA_DIR}/ota_nv_boot_control.conf" "${OTA_BASE_DIR_TMP}"/
}

function update_entry_in_customer_conf()
{
	# Update entry in customer.conf
	local entry_name="${1}"
	local entry_value="${2}"
	local found=

	if [ ! -f "${OTA_CUSTOMER_CONF}" ]; then
		echo "Customer configuration fie ${OTA_CUSTOMER_CONF} is not found"
		exit 1
	fi

	found="$(grep "${1}=" < "${OTA_CUSTOMER_CONF}")"
	if [ "${found}" == "" ]; then
		echo "ERROR: the entry ${entry_name} is not found in ${OTA_CUSTOMER_CONF}"
		exit 1
	fi
	echo "Update the value of ${entry_name} to ${entry_value}"
	echo "sed -i s/^${entry_name}=.*/${entry_name}=${entry_value}/g ${OTA_CUSTOMER_CONF}"
	sed -i "s/^${entry_name}=.*/${entry_name}=${entry_value}/g" "${OTA_CUSTOMER_CONF}"
}

function generate_package()
{
	# OTA payload package for R28.x to R32-ToT:
	# ota_payload_package.tar.gz
	# |-- layout_change
	# |-- nv-l4t-bootloader-config.sh
	# |-- ota_nv_boot_control.conf
	# |-- base_version
	# |-- board_name
	# |-- nv_ota_customer.conf
	# |-- nv_ota_preserve_data.sh
	# |-- nv_ota_rootfs_updater.sh
	# |-- ota_package.tar.sha1sum
	# |-- ota_package.tar
	#     |-- l4t_update_partitions.sh
	#     |-- ota_backup_files_list.txt
	#     |-- upgradetasklist.txt
	#     |-- version.txt
	#     |-- images-R32i
	#     |-- images-R28-R32i
	#         |-- compatible spec 1 (for example, 3310-B01--)
	#         |-- ...
	#         |-- compatible spec N
	#     |-- images-R32A-R32i
	#         |-- compatible spec 1
	#         |-- ...
	#         |-- compatible spec N
	#     |-- images-R32-ToT
	#         |-- compatible spec 1
	#         |-- ...
	#         |-- compatible spec N
	#     |-- system.img.gz.sha1sum
	#     |-- system.img.gz
	# |-- recovery.img.R28.sha1sum
	# |-- recovery.img.R28
	# |-- recovery.dtb.R28.sha1sum
	# |-- recovery.dtb.R28

	# OTA payload package for R32.x to R32-ToT with layout change:
	# ota_payload_package.tar.gz
	# |-- layout_change
	# |-- nv-l4t-bootloader-config.sh
	# |-- ota_nv_boot_control.conf
	# |-- base_version
	# |-- board_name
	# |-- nv_ota_customer.conf
	# |-- nv_ota_preserve_data.sh
	# |-- nv_ota_rootfs_updater.sh
	# |-- ota_package.tar.sha1sum
	# |-- ota_package.tar
	#     |-- l4t_update_partitions.sh
	#     |-- ota_backup_files_list.txt
	#     |-- upgradetasklist.txt
	#     |-- version.txt
	#     |-- images-R32i
	#     |-- images-R32x-R32i
	#         |-- compatible spec 1 (for example, 2888-400-0001-D.0)
	#         |-- ...
	#         |-- compatible spec N (for example, 2888-400-0001-E.0)
	#     |-- images-R32A-R32i
	#         |-- compatible spec 1
	#         |-- ...
	#         |-- compatible spec N
	#     |-- images-R32-ToT
	#         |-- compatible spec 1
	#         |-- ...
	#         |-- compatible spec N
	#     |-- system.img.gz.sha1sum
	#     |-- system.img.gz
	# |-- recovery.img.${version}.sha1sum
	# |-- recovery.img.${version}
	# |-- recovery.dtb.${version}.sha1sum
	# |-- recovery.dtb.${version}

	# OTA payload package for R32.7 to R35-ToT with layout change:
	# ota_payload_package.tar.gz
	# |-- base_version
	# |-- board_name
	# |-- kernel_bootctrl.bin.normal
	# |-- kernel_bootctrl.bin.update
	# |-- layout_change
	# |-- nv-l4t-bootloader-config.sh
	# |-- nv_ota_check_version.sh
	# |-- nv_ota_common_utils.func
	# |-- nv_ota_customer.conf
	# |-- nv_ota_preserve_data.sh
	# |-- nv_ota_rootfs_updater.sh
	# |-- nv_ota_run_tasks.sh
	# |-- nv_ota_update_all_in_recovery.sh
	# |-- nv_ota_update.sh
	# |-- nv_ota_validate.sh
	# |-- ota_nv_boot_control.conf
	# |-- ota_package.tar.sha1sum
	# |-- ota_package.tar
	#     |-- l4t_update_partitions.sh
	#     |-- ota_backup_files_list.txt
	#     |-- upgradetasklist.txt
	#     |-- version.txt
	#     |-- images-R32i (only included for Xavier NX)
	#     |-- images-R32x-R35i
	#         |-- compatible spec 1 (for example, 2888-400-0001-D.0)
	#         |-- ...
	#         |-- compatible spec N (for example, 2888-400-0001-E.0)
	#     |-- images-R35A-R35i
	#         |-- compatible spec 1
	#         |-- ...
	#         |-- compatible spec N
	#     |-- images-R35-ToT
	#         |-- compatible spec 1
	#         |-- ...
	#         |-- compatible spec N
	#     |-- system.img.gz.sha1sum
	#     |-- system.img.gz

	local version="R32x"

	# Control files
	control_file_orig="upgradetasklist.txt.${TARGET_BOARD}.${version}_to_R35-ToT_emmc"
	control_file="upgradetasklist.txt"
	update_script="l4t_update_partitions.sh"

	# OTA version file
	if [[ "${TARGET_BOARD}" == jetson-xavier-nx-devkit* ]]; then
		version_file_orig="qspi_bootblob_ver.txt"
	else
		version_file_orig="emmc_bootblob_ver.txt"
	fi
	version_file="version.txt"

	# Backup list file
	ota_backup_list_file="ota_backup_files_list.txt"

	# Binaries
	t19x_binary_list=("${BOOTLOADER_DIR}/xusb_sil_rel_fw")

	# Copy binaries
	for bin in "${t19x_binary_list[@]}"; do
		cp "${bin}" "${OTA_BASE_DIR_TMP}/${R35_TOT_IMAGES_DIR}"
	done
	if [ "${LAYOUT_CHANGE}" == 1 ]; then
		if [ "${CHIPID}" == "0x19" ]; then
			for bin in "${t19x_binary_list[@]}"; do
				cp "${bin}" "${OTA_BASE_DIR_TMP}/${R32x_R35i_IMAGES_DIR}"
				cp "${bin}" "${OTA_BASE_DIR_TMP}/${R35A_R35i_IMAGES_DIR}"
			done
		fi
	fi

	# Copy control files
	cp "${OTA_DIR}/${update_script}" "${OTA_BASE_DIR_TMP}"
	cp "${OTA_DIR}/${control_file_orig}" "${OTA_BASE_DIR_TMP}/${control_file}"
	sed -i "s/R32i/${R32i_IMAGES_DIR}/" "${OTA_BASE_DIR_TMP}/${control_file}"
	sed -i "s/R32x-R35i/${R32x_R35i_IMAGES_DIR}/" "${OTA_BASE_DIR_TMP}/${control_file}"
	sed -i "s/R35A-R35i/${R35A_R35i_IMAGES_DIR}/" "${OTA_BASE_DIR_TMP}/${control_file}"
	sed -i "s/R35-ToT/${R35_TOT_IMAGES_DIR}/" "${OTA_BASE_DIR_TMP}/${control_file}"

	# Copy the utilities needed for partition update
	if ! copy_utilities_for_ota_update "${PARTITION_UPDATE_UTILITIES[@]}"; then
		echo "Failed to run \"copy_utilities_for_ota_update\""
		return 1
	fi

	# Copy version file
	cp "${BOOTLOADER_DIR}/${version_file_orig}" "${OTA_BASE_DIR_TMP}/${version_file}"

	# Copy backup list file
	cp "${OTA_DIR}/${ota_backup_list_file}" "${OTA_BASE_DIR_TMP}"

	# Generate package
	pushd "${OTA_BASE_DIR_TMP}" > /dev/null 2>&1
	tar cvf "${OTA_PACKAGE_FILE}" ./* --remove-files
	sha1_chksum="$(sha1sum ${OTA_PACKAGE_FILE} | cut -d\  -f 1)"
	echo -n "${sha1_chksum}" > "${OTA_PACKAGE_SHA1_FILE}"

	# Copy common images
	copy_common_images "${version}"

	# Copy ota tasks listed in customer config file
	copy_ota_tasks

	# echo "1" to layout_change file
	echo -n "1" >"${OTA_BASE_DIR_TMP}/layout_change"

	# Copy nv-l4t-bootloader-config.sh and ota_nv_boot_control.conf
	copy_nv_bootloader_config_files

	# Generate ota_payload_package.tar.gz
	tar zcvf "${OTA_PAYLOAD_PACKAGE_ZIP_FILE}" ./* --remove-files
	mkdir -p "${BOOTLOADER_DIR}/${TARGET_BOARD}"
	mv "${OTA_PAYLOAD_PACKAGE_ZIP_FILE}" "${BOOTLOADER_DIR}/${TARGET_BOARD}/"
	popd > /dev/null 2>&1

	rm "${OTA_BASE_DIR_TMP}" -r

	echo -e "\r\nSUCCESS: generate OTA package for update with layout changed \"${BOOTLOADER_DIR}/${TARGET_BOARD}/${OTA_PAYLOAD_PACKAGE_ZIP_FILE}\""
}

# Generate BUP
function generate_BUP()
{
	local bup_generator="${LINUX_BASE_DIR}"/l4t_generate_soc_bup.sh
	local chipid=
	case ${CHIPID} in
	0x18) chipid=t18x; ;;
	0x19) chipid=t19x; ;;
	*)
		echo "Error: un-supported CHIPID(${CHIPID})"
		exit 1
		;;
	esac;

	local board_spec_entry="${!BOARD_SPECS_ARRAY}"
	local board_name="${TARGET_BOARD}"
	local board_spec_file="${BOARD_SPECS_CONFIG_FILE}"
	local payload_dir="${BOOTLOADER_DIR}"/payloads_"${chipid}"
	pushd "${LINUX_BASE_DIR}" > /dev/null 2>&1
	echo "Generate BUP file by running command: ${bup_generator} -f ${board_spec_file} -e ${board_spec_entry} -b ${board_name} ${chipid}"
	if ! "${bup_generator}" -f "${board_spec_file}" -e "${board_spec_entry}" -b "${board_name}" "${chipid}"; then
		echo "Failed to run \"${bup_generator} -f ${board_spec_file} -e ${board_spec_entry} -b ${board_name} ${chipid}\""
		exit 1
	fi
	popd > /dev/null 2>&1

	echo "Copying bup files from ${payload_dir} into ${OTA_BASE_DIR_TMP}"
	cp "${payload_dir}"/* "${OTA_BASE_DIR_TMP}"/
}

copy_utility()
{
	local src="${1}"
	local dst="${2}"

	if [ ! -x "${src}" ]; then
		echo "The utility ${src} is not valid"
		return 1
	fi

	cp "${src}" "${dst}"
	return 0
}

# Copy utilities from rootfs
copy_utilities_for_ota_update()
{
	local utils=("$@")
	local rootfs_dir="${LINUX_BASE_DIR}/rootfs"
	local src=
	local dst=
	local item=
	for item in "${utils[@]}"
	do
		src="${rootfs_dir}/usr/sbin/${item}"
		dst="${OTA_BASE_DIR_TMP}/${item}"
		if ! copy_utility "${src}" "${dst}"; then
			echo "Failed to run \"copy_utility ${src} ${dst}\""
			return 1
		fi
	done
	return 0
}

# Generate OTA payload package for the case of updating without layout change
function generate_package_for_no_layout_change()
{
	# OTA payload package for R32.x to R32-ToT without layout change:
	# ota_payload_package.tar.gz
	# |-- layout_change
	# |-- nv-l4t-bootloader-config.sh
	# |-- ota_nv_boot_control.conf
	# |-- base_version
	# |-- bl_only_payload
	# |-- bl_update_payload
	# |-- board_name
	# |-- kernel_only_payload
	# |-- nvbootctrl
	# |-- nv_bootloader_payload_updater
	# |-- nv_update_engine
	# |-- nv_ota_customer.conf
	# |-- nv_ota_preserve_data.sh
	# |-- nv_ota_rootfs_updater.sh
	# |-- ota_package.tar.sha1sum
	# |-- ota_package.tar
	#     |-- ota_backup_files_list.txt
	#     |-- images-R32-ToT
	#         |-- kernel_bootctrl.bin.update
	#         |-- kernel_bootctrl.bin.normal
	#         |-- recovery.img
	#         |-- recovery.dtb
	# |-- recovery.img.${version}.sha1sum
	# |-- recovery.img.${version}
	# |-- recovery.dtb.${version}.sha1sum
	# |-- recovery.dtb.${version}
	# |-- system.img.gz.sha1sum
	# |-- system.img.gz
	# |-- update_control
	# |-- xusb_only_payload

	local app_system_file="${OTA_BASE_DIR_TMP}/system.img"
	local app_system_sha1_file="${OTA_BASE_DIR_TMP}/system.img.sha1sum"

	if [ "${update_bl_only}" == 0 ]; then
		if [ ! -f "${app_system_file}" ]; then
			echo "Rootfs image ${app_system_file}" is not found
			exit 1
		fi

		if [ ! -f "${app_system_sha1_file}" ]; then
			echo "Rootfs image ${app_system_sha1_file}" is not found
			exit 1
		fi
	fi

	# Remove the files/directories in the "images-R32-ToT" except the
	# following files:
	# recovery image, kernel_bootctrl.bin.update, kernel_bootctrl.bin.normal
	local images_dir=
	local idx_file=
	local tot_recovery_img=
	local tot_recovery_dtb=

	# Get the name of recovery image
	images_dir=$(get_last_interm_dir "${OTA_BASE_DIR_TMP}")
	if [ ! -d "${images_dir}" ]; then
		echo "The directory ${images_dir} is not found"
		exit 1
	fi
	idx_file="${images_dir}/${BOARD_SPEC_NAME}/flash.idx"
	tot_recovery_img="$(grep "recovery," <"${idx_file}" | cut -d, -f 5 | sed 's/^ //g' -)"
	if [ ! -f "${images_dir}/${tot_recovery_img}" ]; then
		echo "The file ${images_dir}/${tot_recovery_img} is not found"
		exit 1
	fi
	tot_recovery_dtb="$(grep "recovery-dtb," <"${idx_file}" | cut -d, -f 5 | sed 's/^ //g' -)"
	if [ ! -f "${images_dir}/${tot_recovery_dtb}" ]; then
		echo "The file ${images_dir}/${tot_recovery_dtb} is not found"
		exit 1
	fi
	pushd "${images_dir}" > /dev/null 2>&1
	mv "${tot_recovery_img}" "recovery.img"
	mv "${tot_recovery_dtb}" "recovery.dtb"
	echo "find . -maxdepth 1 ! -name \"recovery.*\" -exec rm -rf {} \;"
	if ! find . -maxdepth 1 ! -name "recovery.*" -exec rm -rf "{}" \;; then
		echo "Failed to delete dirs/files in the ${images_dir}"
		return 1
	fi
	rm -Rf "${BOARD_SPEC_NAME}"
	popd > /dev/null 2>&1

	# Copy some necessary files for OTA.
	local ota_backup_list_file="ota_backup_files_list.txt"
	local kernel_bootctrl_normal="kernel_bootctrl.bin.normal"
	local kernel_bootctrl_update="kernel_bootctrl.bin.update"
	# Copy kernel_bootctrl.bin.* files
	cp "${OTA_DIR}/${kernel_bootctrl_normal}" "${images_dir}"/
	cp "${OTA_DIR}/${kernel_bootctrl_update}" "${images_dir}"/
	# Copy backup list file
	cp "${OTA_DIR}/${ota_backup_list_file}" "${OTA_BASE_DIR_TMP}"/

	# Generate "ota_package.tar"
	pushd "${OTA_BASE_DIR_TMP}" > /dev/null 2>&1
	tar cvf "${OTA_PACKAGE_FILE}" ./* --remove-files
	sha1_chksum="$(sha1sum ${OTA_PACKAGE_FILE} | cut -d\  -f 1)"
	echo -n "${sha1_chksum}" > "${OTA_PACKAGE_SHA1_FILE}"

	# For the case of no partiton layout change, the version is "R32_3_4"
	local version="R32_3_4"
	# Copy common images
	copy_common_images "${version}"

	# Copy ota tasks listed in customer config file
	copy_ota_tasks

	# Generate BUP and copy it into the ${OTA_BASE_DIR_TMP}"
	generate_BUP

	# Write "update_control" according to the update_bl_only and
	# update_rfs_only.
	if [ "${update_rfs_only}" == 1 ]; then
		echo "rootfs" > "${OTA_BASE_DIR_TMP}/update_control"
	elif [ "${update_bl_only}" == 1 ]; then
		echo "bootloader" > "${OTA_BASE_DIR_TMP}/update_control"
	else
		echo "bootloader" > "${OTA_BASE_DIR_TMP}/update_control"
		echo "rootfs" >> "${OTA_BASE_DIR_TMP}/update_control"
	fi

	# Copy the utilities that are necessary to update bootloader/rootfs
	# without layout chagne
	if ! copy_utilities_for_ota_update "${BUP_UPDATE_UTILITIES[@]}"; then
		echo "Failed to run \"copy_utilities_for_ota_update\""
		return 1
	fi

	# echo "0" to layout_change file
	echo -n "0" >"${OTA_BASE_DIR_TMP}/layout_change"

	# Copy nv-l4t-bootloader-config.sh and ota_nv_boot_control.conf
	copy_nv_bootloader_config_files

	# Generate ota_payload_package.tar.gz
	tar zcvf "${OTA_PAYLOAD_PACKAGE_ZIP_FILE}" ./* --remove-files
	mkdir -p "${BOOTLOADER_DIR}/${TARGET_BOARD}"
	mv "${OTA_PAYLOAD_PACKAGE_ZIP_FILE}" "${BOOTLOADER_DIR}/${TARGET_BOARD}/"
	popd > /dev/null 2>&1

	rm "${OTA_BASE_DIR_TMP}" -r

	echo -e "\r\nSUCCESS: generate OTA package for update without layout change \"${BOOTLOADER_DIR}/${TARGET_BOARD}/${OTA_PAYLOAD_PACKAGE_ZIP_FILE}\""
}

# Make sure that this script is running in root privilege
USERID=$(id -u)
if [ "${USERID}" -ne 0 ]; then
       echo "Please run this program as root."
       exit 0
fi

if [ $# -lt 2 ];then
	usage
fi

# Skip generating system image
skip_system_img=0
# Update bootloader only for update without layout change
update_bl_only=0
# Update rootfs only for update without layout change
update_rfs_only=0
rootfs_updater=""
rootfs_image=
opstr+="brsu:v:i:o:f:"
while getopts "${opstr}" OPTION; do
	case $OPTION in
	u) PKC_KEY_FILE=${OPTARG}; ;;
	v) SBK_KEY_FILE=${OPTARG}; ;;
	s) skip_system_img=1; ;;
	b) update_bl_only=1; ;;
	r) update_rfs_only=1; ;;
	o) rootfs_updater=${OPTARG}; ;;
	f) rootfs_image=${OPTARG}; ;;
	*)
	   usage
	   ;;
	esac;
done

if [ "${PKC_KEY_FILE}" != "" ] && [ ! -f "${PKC_KEY_FILE}" ];then
	echo "Specified PKC key file \"${PKC_KEY_FILE}\" is not found"
	usage
fi

if [ "${SBK_KEY_FILE}" != "" ]; then
	if [ ! -f "${SBK_KEY_FILE}" ];then
		echo "Specified SBK key file \"${SBK_KEY_FILE}\" is not found"
		usage
	else
		if [ "${PKC_KEY_FILE}" == "" ]; then
			echo "PKC key must be provided once SBK key is specified"
			usage
		fi
	fi
fi

if [ "${update_bl_only}" == 1 ] && [ "${update_rfs_only}" == 1 ]; then
	echo "The option -b is conflicted with option -r"
	usage
fi

if [ -n "${rootfs_updater}" ]; then
	rootfs_updater_name="$(basename "${rootfs_updater}")"
	if [ "${rootfs_updater_name}" == "${NV_ROOTFS_UPDATER}" ]; then
		echo "ERROR: the specified rootfs updater name ${rootfs_updater_name} is reserved by Nvidia"
		usage
	elif [ ! -r "${rootfs_updater}" ]; then
		echo "ERROR: the specified rootfs updater ${rootfs_updater} is not readable"
		usage
	else
		rootfs_updater="$(readlink -f "${rootfs_updater}")"
		if [ ! -x "${rootfs_updater}" ]; then
			chmod a+x "${rootfs_updater}"
		fi
		update_entry_in_customer_conf "ROOTFS_UPDATER" "${rootfs_updater_name}"
	fi
fi

if [ -n "${rootfs_image}" ]; then
	if [ ! -f "${rootfs_image}" ]; then
		echo "ERROR: the specified rootfs image ${rootfs_image} is not readable"
		usage
	else
		rootfs_image="$(readlink -f "${rootfs_image}")"
		# Make sure that the "${SYSTEM_IMAGE_RAW_FILE}" is to be
		# deleted if it exists.
		skip_system_img=0
	fi
fi

if [ -z "${rootfs_image}" ] && \
	[ "${skip_system_img}" == 1 ] && \
	[ ! -f "${SYSTEM_IMAGE_RAW_FILE}" ]; then
	echo "The option -s can not be applied if the ${SYSTEM_IMAGE_RAW_FILE} does not exist"
	usage
fi

nargs=$#
BASE_VERSION="${!nargs}"
nargs=$((nargs-1))
TARGET_BOARD="${!nargs}"

# Check the input target board and base version
# If both of them are valid, return the matched board spec;
# Otherwise, report error and exit
BOARD_SPECS_CONFIG_FILE="${OTA_DIR}/ota_board_specs.conf"
if [ ! -f "${BOARD_SPECS_CONFIG_FILE}" ]; then
	echo "The board specs configuration file ${BOARD_SPECS_CONFIG_FILE} is not found"
	usage
fi
source "${OTA_DIR}/ota_validate_params.sh"
if ! ota_validate_params "${TARGET_BOARD}" "${BASE_VERSION}" "internal" "${BOARD_SPECS_CONFIG_FILE}" "BOARD_SPECS_ARRAY" "CHIPID"; then
	echo "Failed to run \"ota_validate_params ${TARGET_BOARD} ${BASE_VERSION} internal ${BOARD_SPECS_CONFIG_FILE} BOARD_SPECS_ARRAY CHIPID\""
	usage
fi

source "${OTA_DIR}/nv_ota_common.func"
# Check whether layout change exists based on the base version.
if ! is_version_no_layout_change "${BASE_VERSION}"; then
	LAYOUT_CHANGE=1
fi

source "${OTA_DIR}/nv_ota_common_utils.func"

init; echo;
generate_binaries; echo;

# Only generate rootfs image for the two cases:
# 1. Update with layout change
# 2. Rootfs needs to be updated in the case of updating without layout change
if [ "${LAYOUT_CHANGE}" == 1 ] || [ "${update_bl_only}" == 0 ]; then
	tar_systemimg; echo;
fi

# Generate different package according to ${LAYOUT_CHANGE}
if [ "${LAYOUT_CHANGE}" == 0 ]; then
	generate_package_for_no_layout_change; echo;
else
	generate_package; echo;
fi
