#!/bin/bash

# Copyright (c) 2021, NVIDIA CORPORATION.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
#
# This is a script to update rootfs partition. NVIDIA provides this script
# as an exmple for updating rootfs partition. Customers can write their
# own scripts to update rootfs partition by referring to this script
#
# This script is to be renamed as "rootfs_updater.sh" and packed into
# OTA payload package when executing the script "l4t_generate_ota_package.sh"
# without specifying any customer provided script for updating rootfs
# partition.
#
# Usage:
#     sudo nv_ota_rootfs_updater.sh [-p <rootfs partition devnode>] \
#	[-d <OTA work dir>] <rootfs image>
#
#
set -e

_COMMON_TAR_OPTIONS=( --checkpoint=10000 --warning=no-timestamp --numeric-owner )
_ENTRY_TABLE_FILE="opt/ota_package/entry_table"
_ENTRY_TABLE_BACKUP="entry_table.ota_backup"
_EXTLINUX_CONF_BACKUP="extlinux.conf.ota_backup"
_EXTLINUX_CONF_FILE="boot/extlinux/extlinux.conf"
_NV_BOOT_CONTROL_CONF="etc/nv_boot_control.conf"
_NV_BOOT_CONTROL_CONF_BACKUP="ota_nv_boot_control.conf"
_OTA_WORK_DIR=
_ROOTFS_PART_DEV=
_ROOTFS_UPDATE_SUCCESS_FLAG="/tmp/rootfs_update_success"

usage()
{
	echo -ne "Usage: sudo $0 [-p <rootfs partition name or devnode>] [-d <OTA work dir>] <rootfs image>\n"
	echo -ne "\t-p <rootfs partition name>: specify the name or devnode of rootfs partition, such as APP or /dev/mmcblk0p1.\n"
	echo -ne "\t-d <OTA work dir>: specify the path of OTA work directory that OTA payload package is decompressed into.\n"
	echo -ne "\t<rootfs image>: specify the compressed rootfs image.\n"
	echo -ne "Example:\n"
	echo -ne "\tsudo $0 -p APP -d /ota_work /ota_work/system.img\n"
	exit 1
}

back_up_file()
{
	# Back up the specified file from rootfs to OTA work directory
	# Usage:
	#        back_up_file {src} {dst}
	local src="${1}"
	local dst="${2}"

	# If ${dst} exists, return as the file has been
	# backed up.
	# Otherwise, back up the file ${src} to ${dst}
	if [ -f "${dst}" ]; then
		echo "The file ${src} has been backed up to ${dst}"
	elif [ -f "${src}" ]; then
		echo "Backing up ${src} to ${dst}"
		cp -f "${src}" "${dst}"
	else
		echo "The specified file ${src} is not found"
		return 1
	fi
}

back_up_specified_files()
{
	# Back up necessary system files and customer specified files
	# Usage:
	#         back_up_specified_files {work_dir} {root_dir}
	local work_dir="${1}"
	local root_dir="${2}"
	local orig_file=
	local backup_file=

	# Back up extlinux.conf
	orig_file="${root_dir}"/"${_EXTLINUX_CONF_FILE}"
	backup_file="${work_dir}"/"${_EXTLINUX_CONF_BACKUP}"
	echo "back_up_file ${orig_file} ${backup_file}"
	if ! back_up_file "${orig_file}" "${backup_file}"; then
		echo "Failed to run \"back_up_file ${orig_file} ${backup_file}\""
		return 1
	fi

	# Back up the nv_boot_control.conf
	orig_file="${root_dir}"/"${_NV_BOOT_CONTROL_CONF}"
	backup_file="${work_dir}"/"${_NV_BOOT_CONTROL_CONF_BACKUP}"
	echo "back_up_file ${orig_file} ${backup_file}"
	if ! back_up_file "${orig_file}" "${backup_file}"; then
		echo "Failed to run \"back_up_file ${orig_file} ${backup_file}\""
		return 1
	fi

	# Back up customer's files
	echo "ota_backup_customer_files ${work_dir} ${root_dir}"
	if ! ota_backup_customer_files "${work_dir}" "${root_dir}"; then
		echo "Failed to run \"ota_backup_customer_files ${work_dir} ${root_dir}\""
		return 1
	fi
}

restore_file()
{
	# Restore the specified file from OTA work directory to rootfs
	# Usage:
	#     restore_file {src} {dst}
	local src="${1}"
	local dst="${2}"

	if [ ! -f "${src}" ]; then
		echo "The file ${src} is not found"
		return 1
	fi
	echo "Restoring the file ${src} to ${dst}"
	cp -f "${src}" "${dst}"
}

restore_specified_files()
{
	# Restore necessary system files and customer specified files
	# Usage:
	#         restore_specified_files {work_dir} {root_dir}
	local work_dir="${1}"
	local root_dir="${2}"
	local restore_file=
	local target_file=

	# Restore customer's files
	echo "ota_restore_customer_files ${work_dir} ${root_dir}"
	if ! ota_restore_customer_files "${work_dir}" "${root_dir}"; then
		echo "Failed to run \"ota_restore_customer_files ${work_dir} ${root_dir}\""
		return 1
	fi

	# Restore the nv_boot_control.conf from OTA work directory
	restore_file="${work_dir}"/"${_NV_BOOT_CONTROL_CONF_BACKUP}"
	target_file="${root_dir}"/"${_NV_BOOT_CONTROL_CONF}"
	echo "restore_file ${restore_file} ${target_file}"
	if ! restore_file "${restore_file}" "${target_file}"; then
		echo "Failed to run \"restore_file ${restore_file} ${target_file}\""
		return 1
	fi

	# Restore exltinux.conf from OTA work directory
	restore_file="${work_dir}"/"${_EXTLINUX_CONF_BACKUP}"
	target_file="${root_dir}"/"${_EXTLINUX_CONF_FILE}"
	echo "restore_file ${restore_file} ${target_file}"
	if ! restore_file "${restore_file}" "${target_file}"; then
		echo "Failed to run \"restore_file ${restore_file} ${target_file}\""
		return 1
	fi

	# Restore entry table from OTA work directory if it exists
	# If bootloader is successfully updated, the entry table is backed
	# up at ${work_dir}/${_ENTRY_TABLE_BACKUP}. And it should be
	# restored to ${root_dir}/${_ENTRY_TABLE_FILE}
	restore_file="${work_dir}"/"${_ENTRY_TABLE_BACKUP}"
	if [ -f "${restore_file}" ]; then
		target_file="${root_dir}"/"${_ENTRY_TABLE_FILE}"
		echo "restore_file ${restore_file} ${target_file}"
		restore_file "${restore_file}" "${target_file}"
	fi
}

update_rootfs_partition()
{
	# Update rootfs partition in recovery kernel or normal kernel
	# Usage:
	#     update_rootfs_partition {file_image} {rootfs_partition} {work_dir}
	local file_image="${1}"
	local rootfs_partition="${2}"
	local work_dir="${3}"
	local tmp_mnt=/tmp/mnt

	echo "Validate ${file_image}"
	if ! gzip -t "${file_image}" >& /dev/null; then
		echo "The rootfs image ${file_image} is invalid."
		return 1
	fi
	mkdir -p "${tmp_mnt}"

	source "${work_dir}"/nv_ota_preserve_data.sh

	# Mount rootfs partition
	echo "Mounting rootfs partition ${rootfs_partition} ..."
	if ! mount "${rootfs_partition}" "${tmp_mnt}"; then
		echo "Failed to mount rootfs partition ${rootfs_partition}"
		return 1
	fi

	# Back up files into OTA work directory
	echo "Backing up files into ${work_dir}"
	if ! back_up_specified_files "${work_dir}" "${tmp_mnt}"; then
		echo "Failed to \"back_up_specified_files ${work_dir} ${tmp_mnt}\""
		return 1
	fi

	# Delete all the dirs/files in the rootfs partition
	# except the ota package and work directory
	echo "Deleting directories/files in the ${rootfs_partition}"
	echo "Entering ${tmp_mnt}"
	pushd "${tmp_mnt}" > /dev/null 2>&1 || return 1
	echo "find \"${tmp_mnt}\" -maxdepth 1 ! -name \"ota*\" -exec rm -rf {} \;"
	if ! find . -maxdepth 1 ! -name "ota*" -exec rm -rf "{}" \;; then
		echo "Failed to delete dirs/files in the ${rootfs_partition}"
		return 1
	fi
	popd > /dev/null 2>&1  || return 1
	echo "Exiting ${tmp_mnt}"

	# Decompress rootfs image into rootfs partition
	echo "Decompress rootfs image into ${rootfs_partition} ..."
	echo "tar --xattrs -xpf ${file_image} ${_COMMON_TAR_OPTIONS[*]} -C ${tmp_mnt}/"
	if ! tar --xattrs -xpf "${file_image}" "${_COMMON_TAR_OPTIONS[@]}" -C "${tmp_mnt}/"; then
		echo "Failed to decompress rootfs image into ${rootfs_partition}"
		return 1
	fi

	# Restore files from OTA work directory
	echo "Restoring files from ${work_dir}"
	if ! restore_specified_files "${work_dir}" "${tmp_mnt}"; then
		echo "Failed to \"restore_specified_files ${work_dir} ${tmp_mnt}\""
		return 1
	fi

	# Umount rootfs parition
	umount "${tmp_mnt}"
	return 0
}

if [ $# -lt 1 ];then
	usage
fi

nargs=$#
_ROOTFS_IMAGE="${!nargs}"

opstr+="fd:p:"
while getopts "${opstr}" OPTION; do
	case $OPTION in
	d) _OTA_WORK_DIR=${OPTARG}; ;;
	p) _ROOTFS_PART_DEV=${OPTARG}; ;;
	*)
	   usage
	   ;;
	esac;
done

# Check passed-in arguments
if [ ! -f "${_ROOTFS_IMAGE}" ]; then
	echo "Specified rootfs image ${_ROOTFS_IMAGE} is not found"
	exit 1
fi
if [ -n "${_OTA_WORK_DIR}" ] && [ ! -d "${_OTA_WORK_DIR}" ]; then
	echo "Specified OTA work directory ${_OTA_WORK_DIR} is not found"
	exit 1
fi

# Check whether the specified rootfs partition node exist
if [ ! -e "${_ROOTFS_PART_DEV}" ]; then
	echo "Specified rootfs partition ${_ROOTFS_PART_DEV} is not found"
	exit 1
fi

# Update rootfs partition
if [ -f "${_ROOTFS_UPDATE_SUCCESS_FLAG}" ]; then
	rm -f "${_ROOTFS_UPDATE_SUCCESS_FLAG}"
fi
echo "update_rootfs_partition ${_ROOTFS_IMAGE} ${_ROOTFS_PART_DEV} ${_OTA_WORK_DIR}"
if ! update_rootfs_partition "${_ROOTFS_IMAGE}" "${_ROOTFS_PART_DEV}" "${_OTA_WORK_DIR}"; then
	echo "Failed to run \"update_rootfs_partition ${_ROOTFS_IMAGE} ${_ROOTFS_PART_DEV} ${_OTA_WORK_DIR}\""
	exit 1
fi
echo -n "1" >"${_ROOTFS_UPDATE_SUCCESS_FLAG}"
